#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import enum
import os
import shutil
import copy
import tempfile
import re

from . import string, jsonfiles
from .folder import Folder
from .simulation import Simulation
from .maker import Maker, MakerUI
from .events import Events
from .errors import *

class EvaluationMode(enum.Enum):
	'''
	Evaluation mode (evaluate each simulation or a group).
	'''

	EACH = enum.auto()
	GROUP = enum.auto()

class Explorer():
	'''
	Use the Maker to generate simulations and search for particular settings values.

	Parameters
	----------
	simulations_folder : Folder|str
		The simulations folder. Either a `Folder` instance or a path to a folder with a configuration file.

	config_name : str
		Name of the config to use with the Maker.

	generate_only : bool
		`True` to not add the simulations to the manager.
	'''

	def __init__(self, simulations_folder, config_name, *, generate_only = True):
		self._simulations_folder = simulations_folder if type(simulations_folder) is Folder else Folder(simulations_folder)
		self._config_name = config_name

		self._generate_only = generate_only
		self._maker_instance = None

		self.default_simulation = {}

		self._map = None
		self._map_output = None

		self._simulations_dir = None
		self._simulations = None

		self._evaluation = None
		self._evaluation_mode = EvaluationMode.EACH

		self._save_function = None
		self._save_folder = None

		self.events = Events(['stopped', 'map-start', 'map-end', 'map-component-start', 'map-component-progress', 'map-component-end', 'searches-start', 'searches-end', 'search-start', 'search-iteration', 'search-end'])

	def __enter__(self):
		'''
		Context manager to call `close()` at the end.
		'''

		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	@property
	def maker(self):
		'''
		Return the Maker instance.

		Returns
		-------
		maker : Maker
			Instance currently in used.
		'''

		if self._maker_instance is None:
			self._maker_instance = Maker(self._simulations_folder, self._config_name, override_options = {'generate_only': self._generate_only})

		return self._maker_instance

	def close(self):
		'''
		Properly exit the Maker instance.
		'''

		try:
			self._maker_instance.close()

		except AttributeError:
			pass

		else:
			self._maker_instance = None

	@property
	def default_simulation(self):
		'''
		Return the current default settings.

		Returns
		-------
		default_simulation : dict
			The default settings.
		'''

		return self._default_simulation.settings

	@default_simulation.setter
	def default_simulation(self, settings):
		'''
		Set the new default settings.

		Parameters
		----------
		settings : dict
			The new settings to use as default.
		'''

		if not('settings' in settings):
			settings = {'settings': settings}

		self._default_simulation = Simulation(self._simulations_folder, settings)

	@property
	def map(self):
		'''
		Get the current map.

		Returns
		-------
		map : dict
			Description of the map.
		'''

		return self._map

	def _normalizeSettings(self, map_component):
		'''
		Normalize the way the settings are defined in a map component.

		Parameters
		----------
		map_component : dict
			The component to alter.
		'''

		if 'setting' in map_component:
			if not('settings' in map_component):
				map_component['settings'] = map_component['setting']

			del map_component['setting']

		if not(type(map_component['settings']) is list):
			map_component['settings'] = [map_component['settings']]

		map_component['settings'] = [{'set_index': 0, **coords} for coords in map_component['settings']]

		if 'foreach' in map_component:
			self._normalizeSettings(map_component['foreach'])

	@map.setter
	def map(self, new_map):
		'''
		Set the map to follow.

		Parameters
		----------
		new_map : dict
			Description of the map.
		'''

		self._map = new_map

		self._normalizeSettings(self._map)

	@property
	def map_depths(self):
		'''
		Extract the different depth levels from the map.

		Returns
		-------
		depths : dict
			The descriptions of the levels.
		'''

		if self._map is None:
			return None

		depth = 0
		depths = {depth: self._map}

		current_depth = self._map
		while 'foreach' in current_depth:
			depth += 1
			current_depth = current_depth['foreach']
			depths[depth] = current_depth

		return depths

	@property
	def evaluation(self):
		'''
		Get the current evaluation function.

		Returns
		-------
		evaluation : function
			Function used to evaluate the simulations.
		'''

		return self._evaluation

	@evaluation.setter
	def evaluation(self, new_evaluation):
		'''
		Set the evaluation function.

		Parameters
		----------
		new_evaluation : function
			The function to use to evaluate the simulations.
		'''

		self._evaluation = new_evaluation

	@property
	def evaluation_mode(self):
		'''
		Get the current evaluation mode.

		Returns
		-------
		evaluation_mode : EvaluationMode
			Current mode.
		'''

		return self._evaluation_mode

	@evaluation_mode.setter
	def evaluation_mode(self, new_mode):
		'''
		Set the evaluation mode.

		Parameters
		----------
		new_mode : EvaluationMode
			Evaluation mode to use.
		'''

		self._evaluation_mode = new_mode

	@property
	def output(self):
		'''
		Get the output of the map exploration.

		Returns
		-------
		map_output : dict
			Output of the exploration.
		'''

		return self._map_output

	@property
	def save_function(self):
		'''
		Get the function currently used to save a simulation.

		Returns
		-------
		save_function : function
			The function.
		'''

		return self._save_function

	@save_function.setter
	def save_function(self, f):
		'''
		Set the function used to save a simulation.

		Parameters
		----------
		f : function
			The function to use. Use `None` to not save anything.
		'''

		self._save_function = f

	@property
	def save_folder(self):
		'''
		Get the folder currently used to save the simulations.

		Returns
		-------
		save_folder : str
			The folder.
		'''

		return self._save_folder

	@save_folder.setter
	def save_folder(self, folder):
		'''
		Set the folder to use to save the simulations.

		Parameters
		----------
		folder : str
			The folder to use.
		'''

		self._save_folder = folder

	def _saveSimulations(self, simulations):
		'''
		Save some simulations using the saving function and folder.

		Parameters
		----------
		simulations : list
			The simulations to save.

		Returns
		-------
		folder : list
			The folders where the files have been saved.
		'''

		if self._save_function is None or self._save_folder is None:
			return None

		try:
			n = len(os.listdir(self._save_folder))

		except FileNotFoundError:
			n = 0

		folder = os.path.join(self._save_folder, str(n))
		folders = []

		for k, simulation in enumerate(simulations):
			subfolder = os.path.join(folder, str(k))
			os.makedirs(subfolder)

			self._save_function(simulation, subfolder)
			simulation.writeSettingsFile('settings.json', folder = subfolder)

			folders.append(subfolder)

		return folders

	def _setSimulations(self, simulations_settings):
		'''
		Set the current set of simulations to consider, based on the default simulation and the given settings.

		Parameters
		----------
		simulations_settings : list
			List of all settings to alter, for each simulation.
			Each item of the list is a list of dict, each dict being a setting with the following keys:
				* `set`: name of the set the setting belongs to,
				* `set_index` (optional): the index of the set,
				* `name`: the name of the setting,
				* `value`: the value of the setting.
		'''

		self._simulations_dir = tempfile.mkdtemp(prefix = 'hateno-explorer_')

		self._simulations = []
		for k, settings in enumerate(simulations_settings):
			simulation = self._default_simulation.copy()
			simulation['folder'] = os.path.join(self._simulations_dir, str(k))

			for setting in settings:
				simulation.getSetting(setting).value = setting['value']

			self._simulations.append(simulation)

		self._simulations_settings = copy.deepcopy(simulations_settings)

	def _generateSimulations(self, simulations = None):
		'''
		Generate some simulations.

		Parameters
		----------
		simulations : list
			The simulations to generate. If `None`, generate the full set defined by `_setSimulations()`.

		Returns
		-------
		saved_folders : list
			The list of folders where the simulations have been saved.
		'''

		if simulations is None:
			simulations = self._simulations

		self.maker.run(simulations)

		return self._saveSimulations(simulations)

	def _deleteSimulations(self):
		'''
		Delete the current set of simulations.
		'''

		try:
			shutil.rmtree(self._simulations_dir)

		except FileNotFoundError:
			pass

		self._simulations_dir = None
		self._simulations = None
		self._simulations_settings = None

	def _checkStopCondition(self, stop_condition, evaluations):
		'''
		If a stop condition is provided, check if it is true or false.
		Access to specific evaluations is possible with indices (e.g. "[0] > [-1]" to check whether the first evaluation is greater than the latest one).
		By default, the condition is tested against the latest evaluation (e.g. "> 0" is then equivalent to "[-1] > 0").
		If we try to access non-existing indices, or if we expect a certain number of evaluations that is not reached yet, always return `False`.

		Parameters
		----------
		stop_condition : str
			Condition to evaluate.

		evaluations : list
			List of evaluations to consider.

		Returns
		-------
		stop : bool
			Result of the test. Always `False` if there is no stop condition.
		'''

		if stop_condition is None:
			return False

		first_operator_match = re.search(r'([<>]=?|[!=]=|in)', stop_condition.strip())

		if first_operator_match and first_operator_match.start() == 0:
			return string.safeEval(f'{evaluations[-1]} {stop_condition}')

		index_regex = re.compile(r'\[(-?[0-9]+)\]')

		# Retrieve the requested indices and first see if the ones which should be different are really different
		# To do that, we "fix" negative indexes so they should be equal to their positive counterparts
		# If they are still negative, it seems that we don't have enough elements in the list: it will be detected in the next part
		# It is important to add the length of the list and to not use the modulo, to be sure we consider different elements

		requested_indices = set(map(int, index_regex.findall(stop_condition)))
		unique_requested_indices = set(map(lambda k: k if k >= 0 else k + len(evaluations), requested_indices))

		if len(unique_requested_indices) < len(requested_indices):
			return False

		try:
			stop_condition = index_regex.sub(lambda m: str(evaluations[int(m.group(1))]), stop_condition)

		except IndexError:
			return False

		else:
			return string.safeEval(stop_condition)

	def _checkAndStoreStopCondition(self, stop_condition, evaluations, output, depth):
		'''
		Check if a stop condition is verified and store the result.

		Parameters
		----------
		stop_condition : str
			Condition to evaluate.

		evaluations : list
			List of evaluations to consider.

		output : dict
			Dictionary in which the ouput should be stored.

		depth : int
			Depth to use in the output.

		Returns
		-------
		stop : bool
			Result of the test.
		'''

		if stop_condition is None:
			return False

		stop_result = self._checkStopCondition(stop_condition, evaluations)

		if not('stops' in output):
			output['stops'] = []

		output['stops'].append({
			'depth': depth,
			'result': stop_result
		})

		return stop_result

	def _evaluateEach(self, depth, stop_condition = None):
		'''
		Call the evaluation function on each simulation of the current set.

		Parameters
		----------
		depth : int
			Current depth in the exploration tree.

		stop_condition : str
			Condition to stop the evaluation.

		Returns
		-------
		output : list
			Result of the evaluation of each simulation. Each item is a dictionary with keys:
				* `settings`: the altered settings of the simulation,
				* `evaluation`: the result of the evaluation of the simulation.
		'''

		output = []
		evaluations = []

		for simulation, settings in zip(self._simulations, self._simulations_settings):
			folders = self._generateSimulations([simulation])

			evaluation = self._evaluation(simulation)
			evaluations.append(evaluation)

			o = {
				'settings': settings,
				'evaluation': evaluation
			}

			if not(folders is None):
				o['save'] = folders[0]

			output.append(o)

			self.events.trigger('map-component-progress', depth)

			if self._checkAndStoreStopCondition(stop_condition, evaluations, o, depth):
				self.events.trigger('stopped')
				break

		return output

	def _evaluateGroup(self):
		'''
		Call the evaluation function on all simulations in the current set, as a group.

		Returns
		-------
		output : dict
			Result of the evaluation of the group. The dictionary has the following keys:
				* `settings`: the altered settings of all simulations for the group,
				* `evaluation`: the result of the evaluation.
		'''

		saved_folders = self._generateSimulations()

		evaluation = self._evaluation(self._simulations)

		output = {
			'settings': self._simulations_settings,
			'evaluation': evaluation
		}

		if not(saved_folders is None):
			output['save'] = os.path.dirname(saved_folders[0])

		return output

	def _buildValues(self, vdesc):
		'''
		Build a list of values from its description.

		Parameters
		----------
		vdesc : list|dict
			Explicit list, or description. A description is a dictionary with the following keys:
				* `from`: the first value(s),
				* `to`: the last value(s),
				* `n`: the number of values.

		Returns
		-------
		values : list
			The complete list of values.
		'''

		if type(vdesc) is list:
			return vdesc

		if type(vdesc['from']) is list:
			return [
				[
					a + k * (b - a) / (vdesc['n'] - 1)
					for a, b in zip(vdesc['from'], vdesc['to'])
				]
				for k in range(0, vdesc['n'])
			]

		return [
			vdesc['from'] + k * (vdesc['to'] - vdesc['from']) / (vdesc['n'] - 1)
			for k in range(0, vdesc['n'])
		]

	def _buildSettings(self, map_component, additional_settings = []):
		'''
		Build a list of settings with their values from a map component.

		Parameters
		----------
		map_component : dict
			The component to interpret.

		additional_settings : list
			List of settings to alter (that are not listed in this component), with their values.

		Returns
		-------
		settings : list
			Each item is the list of all altered settings with their values for one simulation.
		'''

		return [
			additional_settings + [
				{**coords, 'value': value}
				for coords, value in zip(map_component['settings'], values if type(values) is list else [values])
			]
			for values in self._buildValues(map_component['values'])
		]

	def _mapComponent(self, map_component, current_settings = [], depth = 0):
		'''
		Interpret a component of a map.

		Parameters
		----------
		map_component : dict
			The component to interpret.

		current_settings : list
			List of settings to alter (that are not listed in this component), with their values.

		depth : int
			Current depth in the exploration tree.

		Returns
		-------
		result : list
			A list of dict, each one corresponding to a simulation generated by the component. Each dict contains the keys:
				* `settings`: the list of altered settings,
				* `evaluation`: the result of the evaluation function on this simulation.
		'''

		simulations_settings = self._buildSettings(map_component, current_settings)

		self.events.trigger('map-component-start', depth, map_component, simulations_settings)

		output = []

		if 'foreach' in map_component:
			evaluations = []

			for settings in simulations_settings:
				output += self._mapComponent(map_component['foreach'], settings, depth + 1)
				evaluations.append(output[-1]['evaluation'])

				self.events.trigger('map-component-progress', depth)

				if self._checkAndStoreStopCondition(map_component.get('stop'), evaluations, output[-1], depth):
					self.events.trigger('stopped')
					break

		else:
			self._setSimulations(simulations_settings)

			if self._evaluation_mode == EvaluationMode.EACH:
				output = self._evaluateEach(depth, map_component.get('stop'))

			elif self._evaluation_mode == EvaluationMode.GROUP:
				output = [self._evaluateGroup()]

			self._deleteSimulations()

		self.events.trigger('map-component-end', depth, map_component, simulations_settings)

		return output

	def followMap(self):
		'''
		Follow the map to determine which setting(s) to alter and which values to consider.
		'''

		self.events.trigger('map-start')

		self._map_output = {
			'default_settings': self.default_simulation,
			'map': self._map,
			'evaluations': self._mapComponent(self._map)
		}

		self.events.trigger('map-end')

	def findStops(self, depth, *, get_index = False):
		'''
		Find where the stops of a given depth are verified in the current output.
		Return the settings defined until the given depth. Deeper settings are ignored.

		Parameters
		----------
		depth : int
			Depth to find the stops of.

		get_index : bool
			If `True`, the returned list's items are tuples and their second element is the index at which the settings have been found.

		Raises
		------
		ExplorerDepthNotFoundError
			The depth has not been found in the map.

		ExplorerStopNotFoundError
			There is no stop in the description of the given depth.

		Returns
		-------
		settings_with_stop : list
			Settings that led to the verified stop.
		'''

		if self._map_output is None:
			return None

		depths = self.map_depths

		if not(depth in depths):
			raise ExplorerDepthNotFoundError(depth)

		level = depths[depth]

		if not('stop' in level):
			raise ExplorerStopNotFoundError(depth)

		k_max = sum([len(l['settings']) for d, l in depths.items() if d < depth]) + len(level['settings'])

		settings_with_stop = []
		for i, evaluation in enumerate(self._map_output['evaluations']):
			try:
				stops = evaluation['stops']

			except KeyError:
				pass

			else:
				if list(filter(lambda stop: stop['depth'] == depth and stop['result'], stops)):
					settings = evaluation['settings'] if self._evaluation_mode == EvaluationMode.EACH else evaluation['settings'][0]
					settings_with_stop.append(settings[:k_max] if not(get_index) else (settings[:k_max],i))

		return settings_with_stop

	@property
	def searches(self):
		'''
		Get the iterations details of the latest searches.

		Returns
		--------
		searches : list
			Each item represent a search by a dictionary with the following keys:
				* `settings`: the settings of the upper levels,
				* `iterations`: the details of the iterations for this search.

			Each iteration is a dictionary with the following keys:
				* `interval`: the interval considered at this iteration,
				* `evaluations`: the evaluations values for the bounds of the interval.
		'''

		try:
			return self._searches

		except AttributeError:
			return None

	def _searchIteration(self, depth, interval, evaluations, tolerance, itermax):
		'''
		Iteration of the search.

		Parameters
		----------
		depth : int
			The depth to optimize.

		interval : tuple
			Lower and upper bounds of the current search interval.

		evaluations : tuple
			Values of the evaluations at the bounds of the interval.

		tolerance : float
			Length of the interval at which the precision is considered enough.

		itermax : int
			Maximal number of iterations if the tolerance is not reached.

		Returns
		--------
		interval : tuple
			The latest search interval.
		'''

		a, b = interval

		self._searches[-1]['iterations'].append({
			'interval': interval,
			'evaluations': evaluations
		})

		self.events.trigger('search-iteration')

		if abs(b - a) < tolerance or len(self._searches[-1]['iterations']) > itermax:
			return interval

		c = 0.5 * (a + b)

		level = self.map_depths[depth]
		level['values'] = [c]
		self.followMap()

		eval_c = self._map_output['evaluations'][-1]['evaluation']

		if self._checkStopCondition(level['stop'], [evaluations[0], eval_c]):
			return self._searchIteration(depth, (a, c), (evaluations[0], eval_c), tolerance, itermax)

		else:
			return self._searchIteration(depth, (c, b), (eval_c, evaluations[1]), tolerance, itermax)

	def search(self, depth, tolerance = 1E-5, itermax = 100):
		'''
		Search for the best value to verify a stop.
		There must be only one setting at the given depth, as the search is performed using a dichotomy algorithm.
		The stop must involve the last two evaluations.
		The initial values indicated in the map are used to define the search interval: the solution is assumed to be inside it.
		At first, all the initial values are used until the stop is verified.

		Parameters
		----------
		depth : int
			The depth at which the stop should be verified.

		tolerance : float
			Length of the interval at which the precision is considered enough.

		itermax : int
			Maximal number of iterations if the tolerance is not reached.

		Raises
		------
		ExplorerDepthNotFoundError
			The depth has not been found in the map.

		ExplorerStopNotFoundError
			There is no stop in the description of the given depth.

		ExplorerSearchNoSolutionError
			The stop has not been verified with the initial values.
		'''

		if self._map is None:
			return None

		depths = self.map_depths

		if not(depth in depths):
			raise ExplorerDepthNotFoundError(depth)

		level = depths[depth]

		if not('stop' in level):
			raise ExplorerStopNotFoundError(depth)

		if self._map_output is None:
			self.followMap()

		settings_with_stop = self.findStops(depth, get_index = True)

		if not(settings_with_stop):
			raise ExplorerSearchNoSolutionError()

		# Get the first setting combination that led to a stop
		# The stop has been verified at evaluation k
		# Then, the searched value is between evaluation k-1 and evaluation k
		# We also fix the previous settings to the values allowing a solution

		self.events.trigger('searches-start', len(settings_with_stop))

		self._searches = []
		initial_output = self._map_output

		depth_setting_k0 = len(settings_with_stop[0][0]) - 1

		for stopped, k in settings_with_stop:
			self.events.trigger('search-start')

			self._searches.append({
				'previous_settings': stopped[:-1],
				'iterations': []
			})

			a = (initial_output['evaluations'][k-1]['settings'] if self._evaluation_mode == EvaluationMode.EACH else initial_output['evaluations'][k-1]['settings'][-1])[depth_setting_k0]['value']
			b = (initial_output['evaluations'][k]['settings'] if self._evaluation_mode == EvaluationMode.EACH else initial_output['evaluations'][k]['settings'][-1])[depth_setting_k0]['value']

			i0 = 0
			for d in range(0, depth):
				values = [setting['value'] for setting in stopped[i0:i0+len(depths[d]['settings'])]]
				depths[d]['values'] = values if len(values) == 1 else [values]
				i0 += len(depths[d]['settings'])

			self._searchIteration(depth, (a, b), (initial_output['evaluations'][k-1]['evaluation'], initial_output['evaluations'][k]['evaluation']), tolerance, itermax)

			self.events.trigger('search-end')

		self.events.trigger('searches-end')

class ExplorerUI(MakerUI):
	'''
	UI to show the different steps of the Explorer.

	Parameters
	----------
	explorer : Explorer
		Instance of the Explorer from which the event are triggered.
	'''

	def __init__(self, explorer):
		super().__init__(explorer.maker)

		self._explorer = explorer

		self._explorer_state_line = None
		self._explorer_bar = None

		self._components_lines = {}
		self._components_bars = {}

		self._search_started = False

		self._explorer.events.addListener('stopped', self._stopped)
		self._explorer.events.addListener('map-start', self._mapStart)
		self._explorer.events.addListener('map-end', self._mapEnd)
		self._explorer.events.addListener('map-component-start', self._mapComponentStart)
		self._explorer.events.addListener('map-component-progress', self._mapComponentProgress)
		self._explorer.events.addListener('map-component-end', self._mapComponentEnd)
		self._explorer.events.addListener('searches-start', self._searchesStart)
		self._explorer.events.addListener('searches-end', self._searchesEnd)
		self._explorer.events.addListener('search-start', self._searchStart)
		self._explorer.events.addListener('search-iteration', self._searchIteration)
		self._explorer.events.addListener('search-end', self._searchEnd)

		self._explorer.maker.events.addListener('run-end', self._clearMakerState)

	def _updateExplorerState(self, state):
		'''
		Text line to display the current state of the Explorer.

		Parameters
		----------
		state : str
			State to display.
		'''

		if self._explorer_state_line is None:
			self._explorer_state_line = self.addTextLine(state, position = 0)

		else:
			self._explorer_state_line.text = state

	def _clearMakerState(self, *args, **kwargs):
		'''
		We don't need the Maker state anymore: we erase it.
		'''

		self._clearState()

	def clearState(self):
		'''
		Clear the Explorer state.
		'''

		if self._explorer_state_line is not None:
			self.removeItem(self._explorer_state_line)
			self._explorer_state_line = None

	def _stopped(self):
		'''
		A stop condition is verified.
		'''

		pass

	def _mapStart(self):
		'''
		The following of a map is started.
		'''

		if not(self._search_started):
			self._updateExplorerState('Following a map…')

	def _mapEnd(self):
		'''
		The following of a map has ended.
		'''

		if not(self._search_started):
			self._updateExplorerState('Map followed')

	def _mapComponentStart(self, depth, map_component, simulations_settings):
		'''
		The exploration of a new map component has begun.

		Parameters
		----------
		depth : int
			Depth of the component.

		map_component : dict
			Description of the component.

		simulations_settings : list
			Settings of the simulations in the component.
		'''

		if 'foreach' in map_component or self._explorer._evaluation_mode == EvaluationMode.EACH:
			offset = 1 if self._explorer_bar is None else 2
			self._components_lines[depth] = self.addTextLine(f'Component {depth}…', position = offset + 2*depth)
			self._components_bars[depth] = self.addProgressBar(len(simulations_settings), position = offset + 1 + 2*depth)

	def _mapComponentProgress(self, depth):
		'''
		The exploration of a new map component has progressed.

		Parameters
		----------
		depth : int
			Depth of the component.
		'''

		if depth in self._components_bars:
			self._components_bars[depth].counter += 1

	def _mapComponentEnd(self, depth, map_component, simulations_settings):
		'''
		The exploration of a map component has ended.

		Parameters
		----------
		depth : int
			Depth of the component.

		map_component : dict
			Description of the component.

		simulations_settings : list
			Settings of the simulations in the component.
		'''

		if depth in self._components_lines:
			self.removeItem(self._components_bars[depth])
			del self._components_bars[depth]

			self.removeItem(self._components_lines[depth])
			del self._components_lines[depth]

	def _searchesStart(self, n_searches):
		'''
		A search for a best value has begun.

		Parameters
		----------
		n_searches : int
			Number of searches that will be performed.
		'''

		self._search_started = True
		self._updateExplorerState('Searching for the best value…')

		if n_searches > 1:
			self._explorer_bar = self.addProgressBar(n_searches, position = 1)

	def _searchesEnd(self):
		'''
		A search for a best value has ended.
		'''

		self._updateExplorerState('Search finished')

		if self._explorer_bar is not None:
			self.removeItem(self._explorer_bar)
			self._explorer_bar = None

		self._search_started = False

	def _searchStart(self):
		'''
		A search for a best value has begun, with specific previous settings.
		'''

		pass

	def _searchIteration(self):
		'''
		An iteration in the search loop has begun.
		'''

		current_search = self._explorer.searches[-1]
		n_iterations = len(current_search['iterations'])
		interval = current_search['iterations'][-1]['interval']
		self._updateExplorerState(f'Iteration {n_iterations}, interval length: {abs(interval[1] - interval[0])}')

	def _searchEnd(self):
		'''
		A search for a best value has ended, with specific previous settings.
		'''

		if self._explorer_bar is not None:
			self._explorer_bar.counter += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import os
import stat
import shutil
import subprocess
import paramiko

from . import jsonfiles
from .errors import *

class RemoteFolder():
	'''
	Send files to and receive from a remote folder.

	Parameters
	----------
	folder_conf : dict
		Configuration of the remote folder.

	Raises
	------
	FileNotFoundError
		The configuration file does not exist.
	'''

	def __init__(self, folder_conf):
		self._configuration = folder_conf
		self._local = (self._configuration['host'] == 'local')

	def __enter__(self):
		'''
		Context manager to call `open()` and `close()` automatically.
		'''

		self.open()
		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	def open(self):
		'''
		Open the connection.
		'''

		if self._local:
			self._sftp = LocalSFTP()

		else:
			self._ssh = paramiko.SSHClient()
			self._ssh.load_system_host_keys()
			self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

			connect_params = {'username': self._configuration['user']}

			try:
				connect_params['port'] = self._configuration['port']

			except KeyError:
				pass

			self._ssh.connect(self._configuration['host'], **connect_params)

			self._sftp = self._ssh.open_sftp()

		if 'working_directory' in self._configuration:
			self._sftp.chdir(self._configuration['working_directory'])

	def close(self):
		'''
		Close the connection.
		'''

		try:
			self._ssh.close()

		except AttributeError:
			pass

	def execute(self, cmd):
		'''
		Remotely execute a command from the working directory.

		Parameters
		----------
		cmd : str
			Command to execute.

		Returns
		-------
		output : paramiko.ChannelFile
			Output of the command (file-like object).
		'''

		if 'working_directory' in self._configuration:
			cmd = f'cd {self._configuration["working_directory"]}; {cmd}'

		if self._local:
			p = subprocess.run(cmd, shell = True, stdout = subprocess.PIPE)

			stdout = io.StringIO(p.stdout.decode())
			return stdout

		else:
			stdin, stdout, stderr = self._ssh.exec_command(cmd)
			return stdout

	def copyChmodToRemote(self, filename, remote_path):
		'''
		Change the chmod of a remote file to reflect a local one.

		Parameters
		----------
		filename : str
			Name of the file to use to copy the chmod.

		remote_path : str
			Remote path to alter.
		'''

		self._sftp.chmod(remote_path, os.stat(filename).st_mode & 0o777)

	def copyChmodToLocal(self, remote_path, filename):
		'''
		Change the chmod of a local file to reflect a remote one.

		Parameters
		----------
		remote_path : str
			Name of the file to use to copy the chmod.

		filename : str
			Local path to alter.
		'''

		os.chmod(filename, self._sftp.stat(remote_path).st_mode & 0o777)

	def sendFile(self, filename, remote_path = None, *, copy_permissions = True, delete = False):
		'''
		Send a file.

		Parameters
		----------
		filename : str
			Name of the file to send.

		remote_path : str
			Path of the remote file to write.

		copy_permissions : boolean
			`True` to copy the chmod from the local file.

		delete : boolean
			`True` to delete the local file, once sent.

		Returns
		-------
		remote_path : str
			Remote path of the sent file.
		'''

		if not(remote_path):
			remote_path = os.path.basename(filename)

		self._sftp.put(filename, remote_path)

		if copy_permissions:
			self.copyChmodToRemote(filename, remote_path)

		if delete:
			os.unlink(filename)

		return remote_path

	def receiveFile(self, remote_path, filename = None, *, copy_permissions = True, delete = False):
		'''
		Receive (download) a file.

		Parameters
		----------
		remote_path : str
			Path of the remote file to receive.

		filename : str
			Name of the file to create.

		copy_permissions : boolean
			`True` to copy the chmod from the remote file.

		delete : boolean
			`True` to delete the remote file.

		Raises
		------
		RemotePathNotFoundError
			The remote file does not exist.

		Returns
		-------
		filename : str
			Path of the received file.
		'''

		if not(filename):
			filename = os.path.basename(remote_path)

		try:
			self._sftp.get(remote_path, filename)

		except FileNotFoundError:
			raise RemotePathNotFoundError(remote_path)

		if copy_permissions:
			self.copyChmodToLocal(remote_path, filename)

		if delete:
			self._sftp.remove(remote_path)

		return filename

	def getFileContents(self, remote_path):
		'''
		Retrieve the content of a remote file.

		Parameters
		----------
		remote_path : str
			Path of the remote file to read.

		Returns
		-------
		content : str
			Content of the file, as a string.
		'''

		with self._sftp.open(remote_path, 'r') as f:
			return f.read()

	def putFileContents(self, remote_path, content):
		'''
		Write the content of a remote file.

		Parameters
		----------
		remote_path : str
			Path of the remote file to write.

		content : str
			Content to write.
		'''

		with self._sftp.open(remote_path, 'w') as f:
			f.write(content)

	def appendToFile(self, remote_path, content):
		'''
		Append a string to a remote file.

		Parameters
		----------
		remote_path : str
			Path of the remote file to write into.

		content : str
			Content to append.
		'''

		with self._sftp.open(remote_path, 'a') as f:
			f.write(content)

	def makedirs(self, directory):
		'''
		Recursively create a directory.

		Parameters
		----------
		directory : str
			Path to create.
		'''

		try:
			self._sftp.mkdir(directory)

		except FileNotFoundError:
			self.makedirs(os.path.split(os.path.normpath(directory))[0])
			self._sftp.mkdir(directory)

	def sendDir(self, directory, remote_path = None, *, copy_permissions = True, delete = False, empty_dest = False):
		'''
		Send a directory.

		Parameters
		----------
		directory : str
			Name of the directory to send.

		remote_path : str
			Path of the remote directory to create.

		copy_permissions : boolean
			`True` to copy the chmod from the local directory.

		delete : boolean
			`True` to delete the local directory, once sent.

		empty_dest : boolean
			`True` to ensure the destination folder is empty.

		Returns
		-------
		remote_path : str
			Remote path of the sent directory.
		'''

		if not(remote_path):
			remote_path = os.path.basename(os.path.normpath(directory))

		try:
			entries = self._sftp.listdir(remote_path)

			if empty_dest and entries:
				self.deleteRemote([os.path.join(remote_path, e) for e in entries])

		except FileNotFoundError:
			self.makedirs(remote_path)

		if copy_permissions:
			self.copyChmodToRemote(directory, remote_path)

		for entry in [(entry, os.path.join(directory, entry)) for entry in os.listdir(directory)]:
			(self.sendDir if os.path.isdir(entry[1]) else self.sendFile)(entry[1], os.path.join(remote_path, entry[0]), copy_permissions = copy_permissions, delete = delete)

		if delete:
			os.rmdir(directory)

		return remote_path

	def receiveDir(self, remote_path, directory = None, *, copy_permissions = True, delete = False, empty_dest = False):
		'''
		Receive (download) a directory.

		Parameters
		----------
		remote_path : str
			Path of the remote directory to receive.

		directory : str
			Name of the directory to create.

		copy_permissions : boolean
			`True` to copy the chmod from the remote directory.

		delete : boolean
			`True` to delete the remote directory.

		empty_dest : boolean
			`True` to ensure the destination folder is empty.

		Raises
		------
		RemotePathNotFoundError
			The remote directory does not exist.

		Returns
		-------
		directory : str
			Local path of the received directory.
		'''

		try:
			stats = self._sftp.stat(remote_path)

		except FileNotFoundError:
			raise RemotePathNotFoundError(remote_path)

		if not(directory):
			directory = os.path.basename(os.path.normpath(remote_path))

		try:
			entries = os.listdir(directory)

			if empty_dest and entries:
				self.deleteLocal([os.path.join(directory, e) for e in entries])

		except FileNotFoundError:
			os.makedirs(directory)

		if copy_permissions:
			self.copyChmodToLocal(remote_path, directory)

		for entry in [(entry, os.path.join(remote_path, entry)) for entry in self._sftp.listdir(remote_path)]:
			(self.receiveDir if stat.S_ISDIR(self._sftp.stat(entry[1]).st_mode) else self.receiveFile)(entry[1], os.path.join(directory, entry[0]), copy_permissions = copy_permissions, delete = delete)

		if delete:
			self._sftp.rmdir(remote_path)

		return directory

	def send(self, local_path, remote_path = None, *, copy_permissions = True, delete = False, empty_dest = False):
		'''
		Send a file or a directory.

		Parameters
		----------
		local_path : str
			Path of the file/directory to send.

		remote_path : str
			Path of the remote file/directory to create.

		copy_permissions : boolean
			`True` to copy the chmod from the local file/directory.

		delete : boolean
			`True` to delete the local file/directory, once sent.

		empty_dest : boolean
			`True` to ensure the destination folder is empty in the case of a directory.

		Returns
		-------
		remote_path : str
			Remote path of the sent file/directory.
		'''

		kwargs = {
			'copy_permissions': copy_permissions,
			'delete': delete
		}

		send = self.sendFile

		if os.path.isdir(local_path):
			send = self.sendDir
			kwargs['empty_dest'] = empty_dest

		return send(local_path, remote_path, **kwargs)

	def receive(self, remote_path, local_path = None, *, copy_permissions = True, delete = False, empty_dest = False):
		'''
		Receive (download) a file/directory.

		Parameters
		----------
		remote_path : str
			Path of the remote file/directory to receive.

		local_path : str
			Name of the file/directory to create.

		copy_permissions : boolean
			`True` to copy the chmod from the remote file/directory.

		delete : boolean
			`True` to delete the remote file/directory.

		empty_dest : boolean
			`True` to ensure the destination folder is empty in the case of a directory.

		Returns
		-------
		local_path : str
			Local path of the received file/directory.
		'''

		kwargs = {
			'copy_permissions': copy_permissions,
			'delete': delete
		}

		receive = self.receiveFile

		if stat.S_ISDIR(self._sftp.stat(remote_path).st_mode):
			receive = self.receiveDir
			kwargs['empty_dest'] = empty_dest

		return receive(remote_path, local_path, **kwargs)

	def deleteRemote(self, entries):
		'''
		Recursively delete some remote entries.

		Parameters
		----------
		entries : list
			List of paths to delete.
		'''

		for entry in entries:
			if stat.S_ISDIR(self._sftp.stat(entry).st_mode):
				self.deleteRemote([os.path.join(entry, e) for e in self._sftp.listdir(entry)])
				self._sftp.rmdir(entry)

			else:
				self._sftp.remove(entry)

	def deleteLocal(self, entries):
		'''
		Recursively delete some local entries.

		Parameters
		----------
		entries : list
			List of paths to delete.
		'''

		for entry in entries:
			(shutil.rmtree if os.path.isdir(entry) else os.unlink)(entry)

class LocalSFTP():
	'''
	Implement some actions on local files, with the same methods names than paramiko.SFTP.

	Parameters
	----------
	wd : str
		Working directory (base directory to use for the files).
	'''

	def __init__(self, wd = '.'):
		self._wd = wd

	def chdir(self, wd):
		'''
		Change the working directory.

		Parameters
		----------
		wd : str
			Working directory to use.
		'''

		self._wd = wd

	def path(self, path):
		'''
		Prepend a path with the working directory.

		Parameters
		----------
		path : str
			Path to prepend.

		Returns
		-------
		complete_path : str
			The complete path, prepended.
		'''

		return os.path.join(self._wd, path)

	def put(self, local_path, remote_path):
		'''
		Copy a file into the working directory.

		Parameters
		----------
		local_path : str
			Path to the file to copy.

		remote_path : str
			Path of the copied file.
		'''

		shutil.copy(local_path, self.path(remote_path))

	def get(self, remote_path, local_path):
		'''
		Copy a file from the working directory.

		Parameters
		----------
		remote_path : str
			Path to the file to copy.

		local_path : str
			Path of the copied file.
		'''

		shutil.copy(self.path(remote_path), local_path)

	def remove(self, path):
		'''
		Remove a file.

		Parameters
		----------
		path : str
			Path to the file to remove.
		'''

		os.unlink(self.path(path))

	def mkdir(self, dir):
		'''
		Create a directory.

		Parameters
		----------
		dir : str
			Path to the directory to create.
		'''

		os.mkdir(self.path(dir))

	def rmdir(self, dir):
		'''
		Remove a directory.

		Parameters
		----------
		dir : str
			Path to the directory to remove.
		'''

		os.rmdir(self.path(dir))

	def stat(self, path):
		'''
		Get some numbers about a file.

		Parameters
		----------
		path : str
			Path of the file.

		Returns
		-------
		res : os.stat_result
			Result of the stat function.
		'''

		return os.stat(self.path(path))

	def chmod(self, path, mode):
		'''
		Change the mode of a file.

		Parameters
		----------
		path : str
			Path to the file to alter.

		mode : int
			New mode to apply.
		'''

		os.chmod(self.path(path), mode)

	def listdir(self, dir):
		'''
		List a directory.

		Parameters
		----------
		dir : str
			Path of the directory to list.

		Returns
		-------
		content : list
			Entries contained in the directory.
		'''

		return os.listdir(self.path(dir))

	def open(self, path, mode):
		'''
		Open a file.

		Parameters
		----------
		path : str
			Path to the file to open.

		mode : str
			Opening mode.

		Returns
		-------
		fid : TextIOWrapper
			Opened file.
		'''

		return open(self.path(path), mode)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
In this file, the default settings namers are defined.
A namer must admit the following signature.
Convention: prefix the name of the function by `namer_`.

Parameters
----------
name : str
	The name of the setting.

local_index : int
	The index of the setting, inside its set.

local_total : int
	The total number the setting has been used inside its set.

global_index : int
	The index of the setting, among all sets.

global_total : int
	The total number the setting has been used among all sets.

Returns
-------
display_name : str
	The name to use when launching the simulation.
'''

def namer_appendLocalIndex(name, local_index, local_total, global_index, global_total, separator = '-', only_if_multiple = False):
	'''
	Append the local index to the name.

	Parameters
	----------
	only_if_multiple : bool
		Append the index only if the setting is used more than once (locally).

	separator : str
		Separator between the name and the index.
	'''

	if only_if_multiple and local_total <= 1:
		return name

	return name + separator + str(local_index)

def namer_appendGlobalIndex(name, local_index, local_total, global_index, global_total, separator = '-', only_if_multiple = False):
	'''
	Append the local index to the name.

	Parameters
	----------
	only_if_multiple : bool
		Append the index only if the setting is used more than once (globally).

	separator : str
		Separator between the name and the index.
	'''

	if only_if_multiple and global_total <= 1:
		return name

	return name + separator + str(global_index)

def namer_suffix(name, local_index, local_total, global_index, global_total, suffix = ''):
	'''
	Append a string to the name.

	Parameters
	----------
	suffix : str
		The string to append.
	'''

	return name + suffix

def namer_prefix(name, local_index, local_total, global_index, global_total, prefix = ''):
	'''
	Prepend a string to the name.

	Parameters
	----------
	prefix : str
		The string to prepend.
	'''

	return prefix + name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class Error(Exception):
	'''
	Base class for exceptions occurring in the manager.
	'''

	pass

class OperationNotAllowed(Error):
	'''
	Exception raised when we try to perform an operation that is not allowed (e.g. write something in read only mode).
	'''

	pass

class FCollectionCategoryNotFoundError(Error):
	'''
	Exception raised when we try to access a non-existing category of an FCollection.

	Parameters
	----------
	category : str
		Name of the category.
	'''

	def __init__(self, category):
		self.category = category

class FCollectionFunctionNotFoundError(Error):
	'''
	Exception raised when we try to access a non-existing function of an FCollection.

	Parameters
	----------
	fname : str
		Name of the function.
	'''

	def __init__(self, fname):
		self.fname = fname

class InvalidFilterRegexError(Error):
	'''
	Exception raised when we define the filter regex of an FCollection without the required groups.

	Parameters
	----------
	regex : str
		The invalid regex.
	'''

	def __init__(self, regex):
		self.regex = regex

class SettingTagNotRecognizedError(Error):
	'''
	Exception raised when we try to analyse a setting tag but it fails.

	Parameters
	----------
	setting_tag : str
		The invalid tag.
	'''

	def __init__(self, setting_tag):
		self.setting_tag = setting_tag

class JobAlreadyExistingError(Error):
	'''
	Exception raised when we try to add a job to manage that already exists.

	Parameters
	----------
	job_name : str
		Name of the job.
	'''

	def __init__(self, job_name):
		self.job_name = job_name

class JobNotFoundError(Error):
	'''
	Exception raised when we try to access a non existing job.

	Parameters
	----------
	job_name : str
		Name of the job.
	'''

	def __init__(self, job_name):
		self.job_name = job_name

class JobStateNotFoundError(Error):
	'''
	Exception raised when we try to read the state of a job from a dict, without the right key.
	'''

	pass

class UnknownJobStateError(Error):
	'''
	Exception raised when we try to use an unknown job state.

	Parameters
	----------
	state : str
		Name of the unknown state.
	'''

	def __init__(self, state):
		self.state = state

class ManagerAlreadyRunningError(Error):
	'''
	Exception raised when a instance of the Manager is created while another is still running.
	'''

	pass

class SettingsSetNotFoundError(Error):
	'''
	Exception raised when a settings set has not been found.

	Parameters
	----------
	set_name : str
		Name of the set.
	'''

	def __init__(self, set_name):
		self.set_name = set_name

class SettingNotFoundError(Error):
	'''
	Exception raise when a setting has not been found in a given set.

	Parameters
	----------
	set_name : str
		Name of the set.

	setting_name : str
		Name of the setting
	'''

	def __init__(self, set_name, setting_name):
		self.set_name = set_name
		self.setting_name = setting_name

class SimulationFolderNotFoundError(Error):
	'''
	Exception raised when we try to add a simulation to the manager, but the indicated folder does not exist.

	Parameters
	----------
	folder : str
		The folder which has not been found.
	'''

	def __init__(self, folder):
		self.folder = folder

class CheckersCategoryNotFoundError(Error):
	'''
	Exception raised when we try to access a checkers category which does not exist.

	Parameters
	----------
	category : str
		The category which has not been found.
	'''

	def __init__(self, category):
		self.category = category

class CheckerNotFoundError(Error):
	'''
	Exception raised when we try to access a checker which does not exist.

	Parameters
	----------
	category : str
		The checker's category.

	checker_name : str
		The name of the checker which has not been found.
	'''

	def __init__(self, category, checker_name):
		self.category = category
		self.checker_name = checker_name

class VariableGeneratorNotFoundError(Error):
	'''
	Exception raised when we try to access a variable generator which does not exist.

	Parameters
	----------
	generator_name : str
		Name of the generator which has not been found.
	'''

	def __init__(self, generator_name):
		self.generator_name = generator_name

class FixerNotFoundError(Error):
	'''
	Exception raised when we try to access a value fixer which does not exist.

	Parameters
	----------
	fixer_name : str
		Name of the fixer which has not been found.
	'''

	def __init__(self, fixer_name):
		self.fixer_name = fixer_name

class NamerNotFoundError(Error):
	'''
	Exception raised when we try to access a namer which does not exist.

	Parameters
	----------
	namer_name : str
		Name of the namer which has not been found.
	'''

	def __init__(self, namer_name):
		self.namer_name = namer_name

class SimulationIntegrityCheckFailedError(Error):
	'''
	Exception raised when a folder fails to pass an integrity check.

	Parameters
	----------
	folder : str
		The folder which has not been found.
	'''

	def __init__(self, folder):
		self.folder = folder

class SimulationNotFoundError(Error):
	'''
	Exception raised when we look for a non existing simulation.

	Parameters
	----------
	simulation : str
		The name of the simulation which has not been found.
	'''

	def __init__(self, simulation):
		self.simulation = simulation

class SimulationFolderAlreadyExistError(Error):
	'''
	Exception raised when the folder of a simulation already exists.

	Parameters
	----------
	folder : str
		The name of the folder which should not exist.
	'''

	def __init__(self, folder):
		self.folder = folder

class EmptyListError(Error):
	'''
	Exception raised when a specific list is empty.
	'''

	pass

class DestinationFolderExistsError(Error):
	'''
	Exception raised when a destination folder already exists.
	'''

	pass

class ScriptNotFoundError(Error):
	'''
	Exception raised when a given script does not exist.

	Parameters
	----------
	script_coords : dict
		Coordinates of the script which has not been found.
	'''

	def __init__(self, script_coords):
		self.script_coords = script_coords

class RemotePathNotFoundError(Error):
	'''
	Exception raised when we try to access a remote path/directory which does not exist.

	Parameters
	----------
	remote_path : str
		The path.
	'''

	def __init__(self, remote_path):
		self.remote_path = remote_path

class UINonMovableLine(Error):
	'''
	Exception raised when we try to move a line which can't be moved.

	Parameters
	----------
	pos : int
		Position of the line.
	'''

	def __init__(self, pos):
		self.pos = pos

class EventUnknownError(Error):
	'''
	Exception raised when we try to use an unkown event.

	Parameters
	----------
	event : str
		Name of the event.
	'''

	def __init__(self, event):
		self.event = event

class MakerPausedError(Error):
	'''
	Exception raised when the Maker is paused while it should not be.
	'''

	pass

class MakerNotPausedError(Error):
	'''
	Exception raised when the Maker is not paused while it should be.
	'''

	pass

class MakerStateWrongFormatError(Error):
	'''
	Exception raised when we try to read a Maker state that is in the wrong format.
	'''

	pass

class ExplorerDepthNotFoundError(Error):
	'''
	Exception raised when we try to access a non existing depth in a map.

	Parameters
	----------
	depth : int
		The non existing depth.
	'''

	def __init__(self, depth):
		self.depth = depth

class ExplorerStopNotFoundError(Error):
	'''
	Exception raised when we try to access a stop that is not defined.

	Parameters
	----------
	depth : int
		The depth at which the stop was requested.
	'''

	def __init__(self, depth):
		self.depth = depth

class ExplorerSearchNoSolutionError(Error):
	'''
	Exception raised when we search in the wrong interval.
	'''

	pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .fcollection import FCollection
from .errors import *

class Events():
	'''
	Simple implementation of an events handling system.

	Parameters
	----------
	events_names : list
		Names of the events handled by this instance.
	'''

	def __init__(self, events_names):
		self._callbacks = FCollection(categories = events_names)

	def addListener(self, event, f):
		'''
		Add a callback function to a given event.

		Parameters
		----------
		event : str
			Name of the event.

		f : function
			Function to attach.

		Raises
		------
		EventUnknownError
			The event does not exist.
		'''

		try:
			self._callbacks.set(f.__name__, f, category = event)

		except FCollectionCategoryNotFoundError:
			raise EventUnknownError(event)

	def trigger(self, event, *args):
		'''
		Call all functions attached to a given event.

		Parameters
		----------
		event : str
			Name of the event to trigger.

		args : mixed
			Arguments to pass to the callback functions.

		Raises
		------
		EventUnknownError
			The event does not exist.
		'''

		try:
			functions = self._callbacks.getAll(category = event)

		except FCollectionCategoryNotFoundError:
			raise EventUnknownError(event)

		else:
			for f in functions:
				f(*args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
In this file, the default checkers are defined.
A checker must belong into one of the following categories: file checkers, folder checkers, and global checkers.

--

A file checker can be applied to a file. It must have the following signature.
Convention: prefix the name of the function by `file_`.

Parameters
----------
simulation : Simulation
	The simulation which has been generated, as defined by the user.

filename : str
	The name of the file to check (path relative to the simulation's folder).

Returns
-------
success : bool
	`True` if the file successfully passed the test, `False` otherwise.

--

A folder checker can be applied to a folder. It must have the same signature as a file checker (the filename is replaced by the foldername).
Convention: prefix the name of the function by `folder_`.

--

A global checker can be applied to the whole simulation folder. It must have the same signature as a file/folder checker, except for the third argument, described below.
Convention: prefix the name of the function by `global_`.

Parameters
----------
tree : dict
	The list of files and folders names listed in the `output` in the configuration file.
'''

import os
import glob

def file_exists(simulation, filename):
	'''
	File checker.
	Check if at least one file matching the pattern exists.
	'''

	matching_files = [entry for entry in glob.glob(os.path.join(simulation['folder'], filename)) if os.path.isfile(entry)]
	return len(matching_files) > 0

def file_notEmpty(simulation, filename):
	'''
	File checker.
	Check if all files matching the pattern are non empty.
	'''

	nonempty_files = [entry for entry in glob.glob(os.path.join(simulation['folder'], filename)) if os.path.isfile(entry) and os.stat(entry).st_size != 0]
	return len(nonempty_files) > 0

def folder_exists(simulation, foldername):
	'''
	Folder check.
	Check if at least one folder matching the pattern exists.
	'''

	matching_folders = [entry for entry in glob.glob(os.path.join(simulation['folder'], foldername)) if os.path.isdir(entry)]
	return len(matching_folders) > 0

def folder_notEmpty(simulation, foldername):
	'''
	Folder check.
	Check if the folder is not empty.
	'''

	return len(os.listdir(os.path.join(simulation['folder'], foldername))) > 0

def global_noMore(simulation, tree):
	'''
	Global check.
	Check if no other file than the listed ones is present.
	'''

	folder_paths = [(
		list(map(lambda n: os.path.relpath(os.path.join(dirpath, n), simulation['folder']), dirnames)),
		list(map(lambda n: os.path.relpath(os.path.join(dirpath, n), simulation['folder']), filenames))
	) for dirpath, dirnames, filenames in os.walk(simulation['folder'])]

	dirnames, filenames = zip(*folder_paths)
	folder_tree = {
		'folders': sum(dirnames, []),
		'files': sum(filenames, [])
	}

	matching_tree = {
		output_entry: [os.path.relpath(entry, simulation['folder']) for entry in sum([glob.glob(os.path.join(simulation['folder'], pattern)) for pattern in patterns], [])]
		for output_entry, patterns in tree.items()
	}

	return set(folder_tree['folders']) <= set(matching_tree['folders']) and set(folder_tree['files']) <= set(matching_tree['files'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil
import stat
import re
import functools
import copy
from math import ceil
from string import Template

from . import string, jsonfiles
from .errors import *
from .fcollection import FCollection
from .folder import Folder
from .simulation import Simulation
from . import generators as default_generators

class Generator():
	'''
	Generate some scripts to create a set of simulations.
	Initialize the list of simulations to generate.

	Parameters
	----------
	folder : Folder|str
		The folder to manage. Either a `Folder` instance or the path to the folder (used to create a `Folder` instance).
	'''

	def __init__(self, folder):
		self._folder = folder if type(folder) is Folder else Folder(folder)

		self._simulations_to_generate = []

		self._lists_regex_compiled = None
		self._lists_items_regex_compiled = None

		self._variables_generators = None

	@property
	def folder(self):
		'''
		Return the `Folder` instance.

		Returns
		-------
		folder : Folder
			The instance used by the generator.
		'''

		return self._folder

	@property
	def _lists_regex(self):
		'''
		Regex to detect the presence of lists blocks in a skeleton.

		Returns
		-------
		regex : re.Pattern
			The lists regex.
		'''

		if self._lists_regex_compiled is None:
			self._lists_regex_compiled = re.compile(r'^[ \t]*#{3} BEGIN_(?P<tag>[A-Z_]+) #{3}$.+?^(?P<content>.+?)^[ \t]*#{3} END_\1 #{3}$.+?^', flags = re.MULTILINE | re.DOTALL)

		return self._lists_regex_compiled

	@property
	def _lists_items_regex(self):
		'''
		Regex to detect the presence of lists items as variables names.

		Returns
		-------
		regex : re.Pattern
			The lists items regex.
		'''

		if self._lists_items_regex_compiled is None:
			self._lists_items_regex_compiled = re.compile(r'^(?P<list>[A-Z_]+)__(?P<index>[0-9]+)$')

		return self._lists_items_regex_compiled

	@property
	def variables_generators(self):
		'''
		Get the list of available variables generators.

		Returns
		-------
		generators : FCollection
			The collection of generators.
		'''

		if self._variables_generators is None:
			self._variables_generators = FCollection(filter_regex = r'^generator_(?P<name>[A-Za-z0-9_]+)$')
			self._variables_generators.loadFromModule(default_generators)

		return self._variables_generators

	def add(self, simulations):
		'''
		Add simulations to generate.

		Parameters
		----------
		simulations : list|dict|Simulation
			List of simulations to add.
		'''

		if type(simulations) is list:
			for simulation in simulations:
				self._simulations_to_generate.append(Simulation.ensureType(simulation, self._folder))

		else:
			self.add([simulations])

	def clear(self):
		'''
		Clear the list of simulations to generate.
		'''

		self._simulations_to_generate.clear()

	def parse(self, simulations_set = None):
		'''
		Parse a set of simulations to generate the corresponding command lines and other variables.

		Parameters
		----------
		simulations_set : list
			The set of simulations to parse. If `None`, default to the whole list.

		Returns
		-------
		variables : dict
			The list of command lines, and variables corresponding to the simulations' global settings.
		'''

		if simulations_set is None:
			simulations_set = self._simulations_to_generate

		command_lines = [simulation.command_line for simulation in simulations_set]

		globalsettings = self._folder.settings['globalsettings']

		variables = {
			'data_lists': {'COMMAND_LINES': command_lines},
			'data_variables': {'COMMAND_LINES_LENGTH': len(command_lines)}
		}

		for globalsetting in self._folder.settings['globalsettings']:
			name_upper = 'GLOBALSETTING_'+globalsetting['name'].upper()
			variables['data_lists'][name_upper] = [simulation[globalsetting['name']] for simulation in simulations_set]

			if 'generators' in globalsetting:
				for generator_name in globalsetting['generators']:
					try:
						generator = self.variables_generators.get(generator_name)

					except FCollectionFunctionNotFoundError:
						raise VariableGeneratorNotFoundError(generator_name)

					else:
						variables['data_variables'][generator_name.upper()+'_'+name_upper] = functools.reduce(generator, variables['data_lists'][name_upper])

		return variables

	def generateScriptFromSkeleton(self, skeleton_name, output_name, lists, variables, *, make_executable = True):
		'''
		Generate a script from a skeleton, using a given set of command lines.

		Parameters
		----------
		skeleton_name : str
			Name of the skeleton file.

		output_name : str
			Name of the script to write.

		lists : dict
			Lists we can use to loop through.

		variables : dict
			Variables we can use in the whole script template.

		make_executable : boolean
			`True` to add the 'exec' permission to the script.
		'''

		with open(skeleton_name, 'r') as f:
			skeleton = f.read()

		def replaceListBlock(match):
			'''
			Replace a list block by the content of the right list.
			To be called by `re.sub()`.

			Parameters
			----------
			match : re.Match
				Match object corresponding to a list block.

			Returns
			-------
			list_content : str
				The content of the list, formatted as asked.
			'''

			try:
				list_content = ''
				loop_content_template = Template(match.group('content'))

				for index, value in enumerate(lists[match.group('tag')]):
					list_content += loop_content_template.safe_substitute(ITEM_INDEX = index, ITEM_VALUE = value)

				return list_content

			except KeyError:
				return match.group(0)

		script_content = self._lists_regex.sub(replaceListBlock, skeleton)

		variables = copy.deepcopy(variables)
		script_content_template = Template(script_content)

		for match in script_content_template.pattern.finditer(script_content_template.template):
			if match.group('braced'):
				submatch = self._lists_items_regex.match(match.group('braced'))

				if submatch:
					variables[submatch.group(0)] = lists[submatch.group('list')][int(submatch.group('index'))]

		script_content = script_content_template.safe_substitute(**variables)

		with open(output_name, 'w') as f:
			f.write(script_content)

		if make_executable:
			os.chmod(output_name, os.stat(output_name).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

	def _createDestinationFolder(self, dest_folder, *, empty_dest = False):
		'''
		Create the folder where the scripts will be stored.

		Parameters
		----------
		dest_folder : str
			Destination folder where scripts should be stored.

		empty_dest : boolean
			If `True` and if the destination folder already exists, empty it before generating the scripts. If `False` the existence of the folder raises an error.

		Raises
		------
		DestinationFolderExistsError
			The destination folder already exists.
		'''

		if os.path.isdir(dest_folder):
			if empty_dest:
				for entry in [os.path.join(dest_folder, e) for e in os.listdir(dest_folder)]:
					(shutil.rmtree if os.path.isdir(entry) else os.unlink)(entry)

			else:
				raise DestinationFolderExistsError()

		else:
			os.makedirs(dest_folder)

	def _loadRecipe(self, config_name):
		'''
		Load the recipe to use to generate the scripts.

		Parameters
		----------
		config_name : str
			Name of the config folder where `generator.json` will be found.
		'''

		self._recipe = jsonfiles.read(os.path.join(self._folder.config_folder, config_name, 'generator.json'))

	def _splitSimulationsList(self):
		'''
		Split the simulations list into subgroups, according to the recipe.

		Returns
		-------
		subgroups : list
			The subgroups of simulations.
		'''

		simulations_per_group = len(self._simulations_to_generate)

		if 'max_simulations' in self._recipe:
			if self._recipe['max_simulations'] > 0:
				simulations_per_group = self._recipe['max_simulations']

			if 'max_subgroups' in self._recipe and len(self._simulations_to_generate) / simulations_per_group > self._recipe['max_subgroups']:
				simulations_per_group = ceil(len(self._simulations_to_generate) / self._recipe['max_subgroups'])

		elif 'max_subgroups' in self._recipe:
			simulations_per_group = ceil(len(self._simulations_to_generate) / self._recipe['max_subgroups'])

		return [self._simulations_to_generate[k:k+simulations_per_group] for k in range(0, len(self._simulations_to_generate), simulations_per_group)]

	def _dataVariables(self):
		'''
		Use the recipe to define the data variables and lists.

		Returns
		-------
		variables : tuple
			A tuple with (in this order) the data variables, and the data lists.
		'''

		data_variables = {}
		data_lists = {}

		if 'data_lists' in self._recipe:
			data_lists.update(self._recipe['data_lists'])

		if 'data_variables' in self._recipe:
			data_variables.update(self._recipe['data_variables'])

		return (data_variables, data_lists)

	def _skeletons(self):
		'''
		Use the recipe to get the list of skeletons.

		Returns
		-------
		skeletons : tuple
			A tuple with (in this order) the subgroups skeletons, and the wholegroup ones.
		'''

		skeletons_folder = os.path.join(self._folder.skeletons_folder, self._recipe['skeletons'])
		skeletons_recipe = jsonfiles.read(os.path.join(skeletons_folder, 'recipe.json'))

		subgroups_skeletons = [os.path.join(skeletons_folder, skeleton) for skeleton in skeletons_recipe['subgroups']]
		wholegroup_skeletons = [os.path.join(skeletons_folder, skeleton) for skeleton in skeletons_recipe['wholegroup']]

		return (subgroups_skeletons, wholegroup_skeletons)

	def _scriptToLaunch(self, generated_scripts, skeletons):
		'''
		Use the recipe to determine which generated should be called to launch the whole thing.

		Parameters
		----------
		generated_scripts : list
			The list of generated scripts where the one to launch will be picked.

		skeletons : list
			Full list of skeletons.

		Returns
		-------
		script_to_launch : str
			The path to the script to launch.
		'''

		skeletons_folder = os.path.join(self._folder.skeletons_folder, self._recipe['skeletons'])
		skeletons_recipe = jsonfiles.read(os.path.join(skeletons_folder, 'recipe.json'))

		option_split = os.path.join(skeletons_folder, skeletons_recipe['launch']).rsplit(':', maxsplit = 2)
		option_split_num = [string.intOrNone(s) for s in option_split]

		cut = max([k for k, n in enumerate(option_split_num) if n is None]) + 1

		script_name = ':'.join(option_split[:cut])
		coords = option_split_num[cut:]
		coords += [-1] * (2 - len(coords))

		possible_skeletons_to_launch = [
			k
			for k, s in enumerate(skeletons)
			if s == script_name
		]

		return generated_scripts[possible_skeletons_to_launch[coords[0]]][coords[1]]['finalpath']

	def generate(self, dest_folder, config_name, *, empty_dest = False, basedir = None):
		'''
		Generate the scripts to launch the simulations by subgroups.

		Parameters
		----------
		dest_folder : str
			Destination folder where scripts should be stored.

		config_name : str
			Name of the config to use.

		empty_dest : boolean
			If `True` and if the destination folder already exists, empty it before generating the scripts. If `False` the existence of the folder raises an error.

		basedir : str
			Path to the "final" directory from which the scripts will be executed.

		Raises
		------
		EmptyListError
			The list of simulations to generate is empty.

		Returns
		-------
		generated_scripts, script_to_launch : tuple
			List of generated scripts, separated: one list per skeleton, in the order they are called. Second output is the "final" path of the script to launch.
		'''

		if not(self._simulations_to_generate):
			raise EmptyListError()

		self._createDestinationFolder(dest_folder, empty_dest = empty_dest)

		self._loadRecipe(config_name)
		simulations_sets = self._splitSimulationsList()
		data_variables, data_lists = self._dataVariables()
		subgroups_skeletons, wholegroup_skeletons = self._skeletons()

		skeletons_calls = []
		generated_scripts = [[]] * (len(subgroups_skeletons) + len(wholegroup_skeletons))

		jobs_ids = [f'job-{k}' for k in range(0, len(simulations_sets))]

		skeletons_calls += [
			{
				'skeleton_name_joiner': f'-{k}.',
				'skeletons': enumerate(subgroups_skeletons),
				'job_id': jobs_ids[k],
				'jobs_ids': jobs_ids,
				**self.parse(simulations_set)
			}
			for k, simulations_set in enumerate(simulations_sets)
		]

		skeletons_calls.append({
			'skeleton_name_joiner': '.',
			'skeletons': [(len(subgroups_skeletons) + j, s) for j, s in enumerate(wholegroup_skeletons)],
			'job_id': '',
			'jobs_ids': jobs_ids,
			**self.parse()
		})

		scripts_basedir = basedir or dest_folder

		for skeletons_call in skeletons_calls:
			data_lists.update(skeletons_call['data_lists'])
			data_variables.update(skeletons_call['data_variables'])

			data_variables['JOB_ID'] = skeletons_call['job_id']
			data_lists['JOBS_IDS'] = skeletons_call['jobs_ids']

			if 'data_variables_cases' in self._recipe:
				for varname, varparams in self._recipe['data_variables_cases'].items():
					vartest = data_variables[varparams['variable']]
					data_variables[varname] = [value for bound, value in zip(varparams['bounds'], varparams['values']) if bound <= vartest][-1]

			for j, skeleton_name in skeletons_call['skeletons']:
				skeleton_basename_parts = os.path.basename(skeleton_name).rsplit('.skeleton.', maxsplit = 1)
				skeleton_tag = re.sub('[^A-Z_]+', '_', skeleton_basename_parts[0].upper())
				script_name = skeletons_call['skeleton_name_joiner'].join(skeleton_basename_parts)
				script_localpath = os.path.join(dest_folder, script_name)
				script_finalpath = os.path.join(scripts_basedir, script_name)

				self.generateScriptFromSkeleton(skeleton_name, script_localpath, data_lists, data_variables)
				data_variables[skeleton_tag] = script_finalpath

				try:
					data_lists[skeleton_tag].append(script_finalpath)

				except KeyError:
					data_lists[skeleton_tag] = [script_finalpath]

				generated_scripts[j].append({'name': script_name, 'localpath': script_localpath, 'finalpath': script_finalpath})

		return (generated_scripts, self._scriptToLaunch(generated_scripts, subgroups_skeletons + wholegroup_skeletons))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import errno
import copy

from . import jsonfiles
from .errors import *
from .fcollection import FCollection
from . import namers as default_namers
from . import fixers as default_fixers

class Folder():
	'''
	Base class for each system needing access to the configuration files of a simulations folder.
	Initialize with the simulations folder and load the settings.

	Parameters
	----------
	folder : str
		The simulations folder. Must contain a settings file.

	Raises
	------
	FileNotFoundError
		No `simulations.conf` file found in the configuration folder.
	'''

	def __init__(self, folder):
		self._folder = folder
		self._conf_folder_path = os.path.join(self._folder, '.hateno')
		self._settings_file = self.confFilePath('simulations.conf')

		if not(os.path.isfile(self._settings_file)):
			raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self._settings_file)

		self._settings = None

		self._namers = None
		self._fixers = None

	@property
	def folder(self):
		'''
		Return the folder's path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._folder

	@property
	def conf_folder(self):
		'''
		Return the configuration folder path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._conf_folder_path

	@property
	def config_folder(self):
		'''
		Return the path to the `config` folder.

		Returns
		-------
		path : str
			The path.
		'''

		return os.path.join(self._conf_folder_path, 'config')

	@property
	def skeletons_folder(self):
		'''
		Return the path to the `skeletons` folder.

		Returns
		-------
		path : str
			The path.
		'''

		return os.path.join(self._conf_folder_path, 'skeletons')

	def confFilePath(self, filename):
		'''
		Return the path to a configuration file, with a given filename.

		Parameters
		----------
		filename : str
			Name of the file.

		Returns
		-------
		path : str
			Path to the file.
		'''

		return os.path.join(self._conf_folder_path, filename)

	@property
	def settings(self):
		'''
		Return the content of the settings file as a dictionary.

		Returns
		-------
		settings : dict
			The folder's settings.
		'''

		if not(self._settings):
			self._settings = jsonfiles.read(self._settings_file)

			if not('namers' in self._settings):
				self._settings['namers'] = []

			if not('fixers' in self._settings):
				self._settings['fixers'] = []

		return self._settings

	@property
	def fixers(self):
		'''
		Get the list of available values fixers.

		Returns
		-------
		fixers : FCollection
			The collection of values fixers.
		'''

		if self._fixers is None:
			self._fixers = FCollection(filter_regex = r'^fixer_(?P<name>[A-Za-z0-9_]+)$')
			self._fixers.loadFromModule(default_fixers)

		return self._fixers

	@property
	def namers(self):
		'''
		Get the list of available namers.

		Returns
		-------
		namers : FCollection
			The collection of namers.
		'''

		if self._namers is None:
			self._namers = FCollection(filter_regex = r'^namer_(?P<name>[A-Za-z0-9_]+)$')
			self._namers.loadFromModule(default_namers)

		return self._namers

	def applyFixers(self, value, *, before = [], after = []):
		'''
		Fix a value to prevent false duplicates (e.g. this prevents to consider `0.0` and `0` as different values).
		Each item of a list of fixers is either a fixer's name or a list beginning with the fixer's name and followed by the arguments to pass to the fixer.

		Parameters
		----------
		value : mixed
			The value to fix.

		before : list
			List of fixers to apply before the global ones.

		after : list
			List of fixers to apply after the global ones.

		Returns
		-------
		fixed : mixed
			The same value, fixed.

		Raises
		------
		FixerNotFoundError
			The fixer's name has not been found.
		'''

		value = copy.deepcopy(value)

		for fixer in before + self.settings['fixers'] + after:
			if not(type(fixer) is list):
				fixer = [fixer]

			try:
				fixer_func = self.fixers.get(fixer[0])

			except FCollectionFunctionNotFoundError:
				raise FixerNotFoundError(fixer[0])

			else:
				value = fixer_func(value, *fixer[1:])

		return value

	def applyNamers(self, name, local_index, local_total, global_index, global_total, *, before = [], after = []):
		'''
		Transform the name of a setting before being used in a simulation.

		Parameters
		----------
		name : str
			The name of the setting to alter.

		local_index : int
			The index of the setting, inside its set.

		local_total : int
			The total number the setting has been used inside its set.

		global_index : int
			The index of the setting, among all sets.

		global_total : int
			The total number the setting has been used among all sets.

		before : list
			List of namers to apply before the global ones.

		after : list
			List of namers to apply after the global ones.

		Returns
		-------
		name : str
			The name to use.

		Raises
		------
		NamerNotFoundError
			The namer's name has not been found.
		'''

		for namer in before + self.settings['namers'] + after:
			if not(type(namer) is list):
				namer = [namer]

			try:
				namer_func = self.namers.get(namer[0])

			except FCollectionFunctionNotFoundError:
				raise NamerNotFoundError(namer[0])

			else:
				name = namer_func(name, local_index, local_total, global_index, global_total, *namer[1:])

		return name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import inspect

from .errors import *

class FCollection():
	'''
	Represent a collection of functions, with possible categories.

	Parameters
	----------
	categories : list
		Names of the categories to use.

	filter_regex : str
		Regex to use to filter the functions when we load them from a module.
	'''

	def __init__(self, *, categories = [], filter_regex = None):
		self._list = {}
		self._use_categories = bool(categories)

		if categories:
			self._list = {cat: {} for cat in categories}

		if filter_regex:
			self.setFilterRegex(filter_regex)

	def _getList(self, category = None):
		'''
		Get the right list to manage.

		Parameters
		----------
		category : str
			Name of the category, if any.

		Raises
		------
		FCollectionCategoryNotFoundError
			The category has not been found.

		Returns
		-------
		list_to_manage : dict
			The list to manage (either the whole list or a category's one).
		'''

		try:
			list_to_manage = self._list[category] if self._use_categories else self._list

		except KeyError:
			raise FCollectionCategoryNotFoundError(category)

		else:
			return list_to_manage

	def set(self, fname, f, *, category = None):
		'''
		Add a function to the collection, or replace an existing one.

		Parameters
		----------
		fname : str
			Name of the function.

		f : function
			Function to store.

		category : str
			Name of the category, if any.
		'''

		self._getList(category)[fname] = f

	def delete(self, fname, *, category = None):
		'''
		Delete a function.

		Parameters
		----------
		fname : str
			Name of the function to delete.

		category : str
			Name of the category, if any.

		Raises
		------
		FCollectionFunctionNotFoundError
			The function has not been found.
		'''

		list_to_manage = self._getList(category)

		try:
			del list_to_manage[fname]

		except KeyError:
			raise FCollectionFunctionNotFoundError(fname)

	def get(self, fname, *, category = None):
		'''
		Get a function.

		Parameters
		----------
		fname : str
			Name of the function to get.

		category : str
			Name of the category, if any.

		Raises
		------
		FCollectionFunctionNotFoundError
			The function has not been found.

		Returns
		-------
		f : function
			The wanted function.
		'''

		list_to_manage = self._getList(category)

		try:
			return list_to_manage[fname]

		except KeyError:
			raise FCollectionFunctionNotFoundError(fname)

	def getAll(self, *, category = None):
		'''
		Get all functions.

		Parameters
		----------
		category : str
			Name of the category, if any.

		Returns
		-------
		functions : list
			The list of stored functions.
		'''

		return list(self._getList(category).values())

	def setFilterRegex(self, filter_regex):
		'''
		Define the filter regex.
		The regex must define a group named `name` which matches the function's name.
		If the collection is categorized, the regex must also define a group named `category`.

		Parameters
		----------
		filter_regex : str
			Filter regex to define.

		Raises
		------
		InvalidFilterRegexError
			The regex does not contain the required groups.
		'''

		regex = re.compile(filter_regex)

		if not('name' in regex.groupindex) or (self._use_categories and not('category' in regex.groupindex)):
			raise InvalidFilterRegexError(filter_regex)

		else:
			self._filter_regex = regex

	def loadFromModule(self, module):
		'''
		Load functions from a given module, according to the filter regex.

		Parameters
		----------
		module : Module
			Module (already loaded) where are defined the functions.
		'''

		for function in inspect.getmembers(module, inspect.isfunction):
			match = self._filter_regex.match(function[0])

			if match:
				category = match.group('category') if self._use_categories else None
				self.set(match.group('name'), function[1], category = category)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil
import stat
import time
import tempfile

from . import string, jsonfiles

from .events import Events
from .folder import Folder
from .simulation import Simulation
from .manager import Manager
from .generator import Generator
from .remote import RemoteFolder
from .jobs import JobsManager, JobState
from .errors import *
from .ui import UI

class Maker():
	'''
	Assemble all components to extract simulations and automatically create them if they don't exist.

	Parameters
	----------
	simulations_folder : Folder|str
		The simulations folder. Must contain a settings file.

	config_name : str
		Name of the config to use.

	override_options : dict
		Options to override.
	'''

	def __init__(self, simulations_folder, config_name, *, override_options = {}):
		self._simulations_folder = simulations_folder if type(simulations_folder) is Folder else Folder(simulations_folder)
		self._config_name = config_name

		self._manager_instance = None
		self._generator_instance = None
		self._remote_folder_instance = None
		self._jobs_manager = JobsManager()

		self._loadOptions(override_options)

		self._simulations_to_extract = []
		self._unknown_simulations = []
		self._jobs_ids = []

		self._remote_scripts_dir = None

		self._corruptions_counter = 0
		self._failures_counter = 0

		self._paused = False
		self._state_attrs = ['simulations_to_extract', 'corruptions_counter', 'failures_counter', 'unknown_simulations', 'jobs_ids', 'remote_scripts_dir']

		self.events = Events(['close-start', 'close-end', 'remote-open-start', 'remote-open-end', 'delete-scripts', 'paused', 'resume', 'run-start', 'run-end', 'extract-start', 'extract-end', 'extract-progress', 'generate-start', 'generate-end', 'wait-start', 'wait-progress', 'wait-end', 'download-start', 'download-progress', 'download-end', 'addition-start', 'addition-progress', 'addition-end'])

	def __enter__(self):
		'''
		Context manager to call `close()` at the end.
		'''

		return self

	def __exit__(self, type, value, traceback):
		'''
		Ensure `close()` is called when exiting the context manager.
		'''

		self.close()

	@property
	def folder(self):
		'''
		Return the `Folder` instance.

		Returns
		-------
		folder : Folder
			The instance used by the maker.
		'''

		return self._simulations_folder

	@property
	def manager(self):
		'''
		Returns the instance of Manager used in the Maker.

		Returns
		-------
		manager : Manager
			Current instance, or a new one if `None`.
		'''

		if not(self._manager_instance):
			self._manager_instance = Manager(self._simulations_folder)

		return self._manager_instance

	@property
	def generator(self):
		'''
		Returns the instance of Generator used in the Maker.

		Returns
		-------
		generator : Generator
			Current instance, or a new one if `None`.
		'''

		if not(self._generator_instance):
			self._generator_instance = Generator(self._simulations_folder)

		return self._generator_instance

	@property
	def _remote_folder(self):
		'''
		Returns the instance of RemoteFolder used in the Maker.

		Returns
		-------
		remote_folder : RemoteFolder
			Current instance, or a new one if `None`.
		'''

		if not(self._remote_folder_instance):
			self._remote_folder_instance = RemoteFolder(jsonfiles.read(os.path.join(self.folder.config_folder, self._config_name, 'folder.json')))

			self.events.trigger('remote-open-start')
			self._remote_folder_instance.open()
			self.events.trigger('remote-open-end')

		return self._remote_folder_instance

	def close(self):
		'''
		Clear all instances of the modules.
		'''

		self.events.trigger('close-start')

		self._generator_instance = None

		try:
			self._manager_instance.close()

		except AttributeError:
			pass

		try:
			self._remote_folder_instance.close()

		except AttributeError:
			pass

		self._remote_folder_instance = None

		self.events.trigger('close-end')

	def _loadOptions(self, override = {}):
		'''
		Load the options of the Maker, stored in the config folder.

		Parameters
		----------
		override : dict
			Options to impose the value of, despite the values in the config folder.
		'''

		self._options = {
			'settings_file': 'settings.json',
			'max_corrupted': -1,
			'max_failures': 0,
			'jobs_states_filename': 'jobs.txt',
			'jobs_output_filename': 'job.out',
			'generate_only': False
		}

		try:
			self._options.update(jsonfiles.read(os.path.join(self.folder.config_folder, self._config_name, 'maker.json')))

		except FileNotFoundError:
			pass

		self._options.update(override)

	@property
	def paused(self):
		'''
		Getter for the paused state.

		Returns
		-------
		paused : bool
			`True` if the Maker has been paused, `False` otherwise.
		'''

		return self._paused

	def pause(self):
		'''
		Pause the Maker.

		Raises
		------
		MakerPausedError
			The Maker is already in paused state.
		'''

		if self._paused:
			raise MakerPausedError()

		self._paused = True
		self.events.trigger('paused')

	def resume(self):
		'''
		Resume after a pause.

		Returns
		-------
		unknown_simulations : list
			List of simulations that failed to be generated. `None` if the Maker has been paused.

		Raises
		------
		MakerNotPausedError
			The Maker is not in paused state.
		'''

		if not(self._paused):
			raise MakerNotPausedError()

		self._paused = False
		self.events.trigger('resume')
		return self.run(self._simulations_to_extract)

	def saveState(self, filename):
		'''
		Save the current state of the Maker when it is paused.

		Parameters
		----------
		filename : str
			Name of the file to use to write the state.

		Raises
		------
		MakerNotPausedError
			The Maker is not in paused state.
		'''

		if not(self._paused):
			raise MakerNotPausedError()

		state = {attr: getattr(self, f'_{attr}') for attr in self._state_attrs}

		jsonfiles.write(state, filename)

	def loadState(self, filename):
		'''
		Load a state.

		Parameters
		----------
		filename : str
			Name of the file to use to read the state.

		Raises
		------
		MakerNotPausedError
			The Maker is not in paused state.

		MakerStateWrongFormatError
			At least one key is missing in the stored state.
		'''

		if not(self._paused):
			raise MakerNotPausedError()

		state = jsonfiles.read(filename)

		try:
			for attr in self._state_attrs:
				setattr(self, f'_{attr}', state[attr])

		except KeyError:
			raise MakerStateWrongFormatError()

	def run(self, simulations, *, corruptions_counter = 0, failures_counter = 0):
		'''
		Main loop, run until all simulations are extracted or some jobs failed.

		Parameters
		----------
		simulations : list
			List of simulations to extract/generate.

		corruptions_counter : int
			Initial value of the corruptions counter.

		failures_counter : int
		 	Initial value of the failures counter.

		Returns
		-------
		unknown_simulations : list
			List of simulations that failed to be generated. `None` if the Maker has been paused.
		'''

		self.events.trigger('run-start')

		self._simulations_to_extract = simulations

		self._corruptions_counter = corruptions_counter
		self._failures_counter = failures_counter

		while self._runLoop():
			pass

		if self.paused:
			return None

		self.events.trigger('run-end', self._unknown_simulations)

		return self._unknown_simulations

	def _runLoop(self):
		'''
		One loop of the `run()` method.

		Returns
		-------
		continue : bool
			`True` to continue the loop, `False` to break it.
		'''

		if not(self._jobs_ids):
			self.extractSimulations()

			if not(self._unknown_simulations):
				return False

			self.generateSimulations()

		try:
			if not(self.waitForJobs()):
				self._failures_counter += 1

		except KeyboardInterrupt:
			self.pause()
			return False

		if not(self.downloadSimulations()):
			self._corruptions_counter += 1

		self.events.trigger('delete-scripts')
		self._remote_folder.deleteRemote([self._remote_scripts_dir])

		return (self._options['max_corrupted'] < 0 or self._corruptions_counter <= self._options['max_corrupted']) and (self._options['max_failures'] < 0 or self._failures_counter <= self._options['max_failures'])

	def extractSimulations(self):
		'''
		Try to extract the simulations.
		'''

		self.events.trigger('extract-start', self._simulations_to_extract)

		self._unknown_simulations = self.manager.batchExtract(self._simulations_to_extract, settings_file = self._options['settings_file'], callback = lambda : self.events.trigger('extract-progress'))

		if self._options['generate_only']:
			self._unknown_simulations = list(filter(lambda simulation: not(os.path.isdir(simulation['folder'])), self._unknown_simulations))

		self.events.trigger('extract-end')

	def generateSimulations(self):
		'''
		Generate the scripts to generate the unknown simulations, and run them.

		Returns
		-------
		jobs_ids : list
			IDs of the jobs to wait.
		'''

		self.events.trigger('generate-start')

		scripts_dir = tempfile.mkdtemp(prefix = 'simulations-scripts_')
		self._remote_scripts_dir = self._remote_folder.sendDir(scripts_dir)

		self.generator.add(self._unknown_simulations)
		generated_scripts, script_to_launch = self.generator.generate(scripts_dir, self._config_name, empty_dest = True, basedir = self._remote_scripts_dir)
		self.generator.clear()

		self._remote_folder.sendDir(scripts_dir, delete = True, empty_dest = True)

		output = self._remote_folder.execute(script_to_launch)
		self._jobs_ids = list(map(lambda l: l.strip(), output.readlines()))

		self.events.trigger('generate-end')

	def waitForJobs(self):
		'''
		Wait for the jobs to finish.

		Returns
		-------
		success : bool
			`True` is all jobs were finished normally, `False` if there was at least one failure.
		'''

		self.events.trigger('wait-start', self._jobs_ids)

		jobs_by_state = {}
		previous_states = {}

		self._jobs_manager.add(*self._jobs_ids, ignore_existing = True)
		self._jobs_manager.linkToFile(self._options['jobs_states_filename'], remote_folder = self._remote_folder)

		while True:
			self._jobs_manager.updateFromFile()
			jobs_by_state = {
				state: self._jobs_manager.getJobsWithStates([JobState[state.upper()]])
				for state in ['waiting', 'running', 'succeed', 'failed']
			}

			if jobs_by_state != previous_states:
				self.events.trigger('wait-progress', jobs_by_state)

				if set([job['name'] for job in jobs_by_state['succeed'] + jobs_by_state['failed']]) == set(self._jobs_ids):
					break

			previous_states = jobs_by_state
			time.sleep(0.5)

		self._jobs_manager.clear()
		self._jobs_ids = []

		self._remote_folder.deleteRemote([self._options['jobs_states_filename']])

		self.events.trigger('wait-end')

		return not(jobs_by_state['failed'])

	def downloadSimulations(self):
		'''
		Download the generated simulations and add them to the manager.

		Returns
		-------
		success : bool
			`True` if all simulations has successfully been downloaded and added, `False` if there has been at least one issue.
		'''

		self.events.trigger('download-start', self._unknown_simulations)

		success = True

		for simulation in self._unknown_simulations:
			simulation = Simulation.ensureType(simulation, self._simulations_folder)

			tmpdir = tempfile.mkdtemp(prefix = 'simulation_')
			try:
				self._remote_folder.receiveDir(simulation['folder'], tmpdir, delete = True)

			except RemotePathNotFoundError:
				pass

			simulation_dest = simulation['folder']
			simulation['folder'] = tmpdir

			if self._options['generate_only']:
				if self.manager.checkIntegrity(simulation):
					destination_path = os.path.dirname(os.path.normpath(simulation_dest))
					if destination_path and not(os.path.isdir(destination_path)):
						os.makedirs(destination_path)

					os.rename(simulation['folder'], simulation_dest)
					simulation['folder'] = simulation_dest

					if self._options['settings_file']:
						simulation.writeSettingsFile(self._options['settings_file'])

				else:
					shutil.rmtree(simulation['folder'])
					success = False

			else:
				try:
					self.manager.add(simulation)
					simulation['folder'] = simulation_dest

				except (SimulationFolderNotFoundError, SimulationIntegrityCheckFailedError):
					success = False

			self.events.trigger('download-progress')

		self.events.trigger('download-end')

		return success

class MakerUI(UI):
	'''
	UI to show the different steps of the Maker.

	Parameters
	----------
	maker : Maker
		Instance of the Maker from which the event are triggered.
	'''

	def __init__(self, maker):
		super().__init__()

		self._maker = maker

		self._state_line = None
		self._main_progress_bar = None

		self._statuses = 'Current statuses: {waiting} waiting, {running} running, {succeed} succeed, {failed} failed'
		self._statuses_line = None

		self._jobs_lines = {}
		self._jobs_bars = {}

		self._maker.events.addListener('close-start', self._closeStart)
		self._maker.events.addListener('close-end', self._closeEnd)
		self._maker.events.addListener('remote-open-start', self._remoteOpenStart)
		self._maker.events.addListener('remote-open-end', self._remoteOpenEnd)
		self._maker.events.addListener('delete-scripts', self._deleteScripts)
		self._maker.events.addListener('paused', self._paused)
		self._maker.events.addListener('resume', self._resume)
		self._maker.events.addListener('run-start', self._runStart)
		self._maker.events.addListener('run-end', self._runEnd)
		self._maker.events.addListener('extract-start', self._extractStart)
		self._maker.events.addListener('extract-progress', self._extractProgress)
		self._maker.events.addListener('extract-end', self._extractEnd)
		self._maker.events.addListener('generate-start', self._generateStart)
		self._maker.events.addListener('generate-end', self._generateEnd)
		self._maker.events.addListener('wait-start', self._waitStart)
		self._maker.events.addListener('wait-progress', self._waitProgress)
		self._maker.events.addListener('wait-end', self._waitEnd)
		self._maker.events.addListener('download-start', self._downloadStart)
		self._maker.events.addListener('download-progress', self._downloadProgress)
		self._maker.events.addListener('download-end', self._downloadEnd)

	def _updateState(self, state):
		'''
		Text line to display the current state of the Maker.

		Parameters
		----------
		state : str
			State to display.
		'''

		if self._state_line is None:
			self._state_line = self.addTextLine(state)

		else:
			self._state_line.text = state

	def _clearState(self):
		'''
		Remove the state line.
		'''

		if not(self._state_line is None):
			self.removeItem(self._state_line)
			self._state_line = None

	def _closeStart(self):
		'''
		Maker starts closing.
		'''

		pass

	def _closeEnd(self):
		'''
		Maker is closed.
		'''

		pass

	def _remoteOpenStart(self):
		'''
		Connection to the RemoteFolder is started.
		'''

		self._updateState('Connection…')

	def _remoteOpenEnd(self):
		'''
		Connected to the RemoteFolder.
		'''

		self._updateState('Connected')

	def _deleteScripts(self):
		'''
		Deletion of the scripts.
		'''

		self._updateState('Deleting the scripts…')

	def _paused(self):
		'''
		The Maker has been paused.
		'''

		# Erase the "^C" due to the keyboard interruption
		print('\r  ', end = '\r')

		if self._main_progress_bar is not None:
			self.removeItem(self._main_progress_bar)
			self._main_progress_bar = None

		if self._statuses_line is not None:
			self.removeItem(self._statuses_line)
			self._statuses_line = None

		for line, bar in zip(self._jobs_lines.values(), self._jobs_bars.values()):
			self.removeItem(line)
			self.removeItem(bar)

		self._jobs_lines.clear()
		self._jobs_bars.clear()

		self._updateState('Paused')

	def _resume(self):
		'''
		Resume after a pause.
		'''

		pass

	def _runStart(self):
		'''
		The run loop just started.
		'''

		self._updateState('Running the Maker…')

	def _runEnd(self, unknown_simulations):
		'''
		The run loop has ended.

		Parameters
		----------
		unknown_simulations : list
			List of simulations that still do not exist.
		'''

		if unknown_simulations:
			self._updateState(string.plural(len(unknown_simulations), 'simulation still does not exist', 'simulations still do not exist'))

		else:
			self._updateState('All simulations have successfully been extracted')

	def _extractStart(self, simulations):
		'''
		Start the extraction of the simulations.

		Parameters
		----------
		simulations : list
			List of the simulations that will be extracted.
		'''

		self._updateState('Extracting the simulations…')
		self._main_progress_bar = self.addProgressBar(len(simulations))

	def _extractProgress(self):
		'''
		A simulation has just been extracted.
		'''

		self._main_progress_bar.counter += 1

	def _extractEnd(self):
		'''
		All simulations have been extracted.
		'''

		self.removeItem(self._main_progress_bar)
		self._main_progress_bar = None
		self._updateState('Simulations extracted')

	def _generateStart(self):
		'''
		Start the generation of the scripts.
		'''

		self._updateState('Generating the scripts…')

	def _generateEnd(self):
		'''
		Scripts are generated.
		'''

		self._updateState('Scripts generated')

	def _waitStart(self, jobs_ids):
		'''
		Start to wait for some jobs.

		Parameters
		----------
		jobs_ids : list
			IDs of the jobs to wait.
		'''

		self._updateState('Waiting for jobs to finish…')
		self._main_progress_bar = self.addProgressBar(len(jobs_ids))
		self._statuses_line = self.addTextLine(self._statuses.format(waiting = 0, running = 0, succeed = 0, failed = 0))

	def _waitProgress(self, jobs_by_state):
		'''
		The state of at least one job has changed.
		Update the global statuses line and progress bar.
		Display a progress bar for each running job.

		Parameters
		----------
		jobs_by_state : dict
			The jobs IDs, sorted by their state.
		'''

		self._statuses_line.text = self._statuses.format(**{state: len(jobs) for state, jobs in jobs_by_state.items()})
		self._main_progress_bar.counter = len(jobs_by_state['succeed'] + jobs_by_state['failed'])

		for job in jobs_by_state['running']:
			if job['total_steps'] > 0:
				if not(job['name'] in self._jobs_lines):
					self._jobs_lines[job['name']] = self.addTextLine(f'Job {job["name"]} running…')
					self._jobs_bars[job['name']] = self.addProgressBar(job['total_steps'])

				self._jobs_bars[job['name']].counter = job['finished_steps']

		for job in jobs_by_state['succeed'] + jobs_by_state['failed']:
			if job['name'] in self._jobs_lines:
				self.removeItem(self._jobs_lines[job['name']])
				del self._jobs_lines[job['name']]

				self.removeItem(self._jobs_bars[job['name']])
				del self._jobs_bars[job['name']]

	def _waitEnd(self):
		'''
		All jobs are finished.
		'''

		self.removeItem(self._statuses_line)
		self._statuses_line = None

		self.removeItem(self._main_progress_bar)
		self._main_progress_bar = None

		self._updateState('Jobs finished')

	def _downloadStart(self, simulations):
		'''
		Start to download and add the simulations.

		Parameters
		----------
		simulations : list
			Simulations that will be downloaded.
		'''

		self._updateState('Downloading the simulations…')
		self._main_progress_bar = self.addProgressBar(len(simulations))

	def _downloadProgress(self):
		'''
		A simulation has just been downloaded and added.
		'''

		self._main_progress_bar.counter += 1

	def _downloadEnd(self):
		'''
		All simulations have been downloaded and added.
		'''

		self.removeItem(self._main_progress_bar)
		self._main_progress_bar = None
		self._updateState('Simulations downloaded')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import importlib.util
import json

def read(filename, *, allow_generator = False):
	'''
	Read a JSON file.

	Parameters
	----------
	filename : str
		Path to the JSON file to read.

	allow_generator : bool
		`True` to allow the use of a Python script as a generator of the object, `False` to allow JSON only.

	Returns
	-------
	obj : dict|list
		The object described in the JSON file.
	'''

	try:
		with open(filename, 'r') as f:
			return json.loads(f.read())

	except json.decoder.JSONDecodeError:
		if not(allow_generator):
			raise

		module_name = 'tmpobjectgenerator'
		spec = importlib.util.spec_from_file_location(module_name, filename)
		module = importlib.util.module_from_spec(spec)
		sys.modules[module_name] = module
		spec.loader.exec_module(module)

		return module.generate()

def write(obj, filename, *, sort_keys = True):
	'''
	Save an object into a JSON file.

	Parameters
	----------
	obj : dict|list
		Object to save.

	filename : str
		Path to the JSON file.

	sort_keys : bool
		`True` to sort the keys before writing the file.
	'''

	with open(filename, 'w') as f:
		json.dump(obj, f, sort_keys = sort_keys, indent = '\t', separators = (',', ': '))
